#include "os.h"
#define HOR_POLARITY	0x01
#define VER_POLARITY	0x02
#define INTERLACED	0x04
#define DOUBLESCAN	0x08


#define HTOTAL_9	1
#define HDEND_9	2
#define HSSTART_9	4


/* Auflsungsstruktur */
typedef struct
{
	WORD	h_total, h_disp, h_sync_strt, h_sync_wid;
	WORD	v_total, v_disp, v_sync_strt, v_sync_wid;
	WORD	planes;
	WORD	mode;
	WORD	freq;
	WORD	flags;
	UWORD	pitch;
} ATI_RES;


#define A860_DELAY_L    0x80
#define CX_CLK	5			/* Index der benutzerdefinierten Frequenz */

static void InitEngine(void);
static void InitHardware(void);
static void	FillColors(char value);
void SetRS2RS3(UCHAR dacRS2RS3);
static VOID	CalcResValues(RESOLUTION *r, ATI_RES *res);
static WORD	ReadBit(UINT value, int count, char sbit, char dbit);
static char	WriteBit(UINT value, UCHAR old_value,
							int count, char sbit, char dbit);
static UWORD GetReg16(UBYTE *adr);
static void	SetFreq(WORD freq, WORD flag);
static void Prog_ICS2595( char entry, unsigned MHz100);
static WORD Prog_STG1703( char entry, unsigned MHz100);
static void delay_26us( void);
static void delay_1ms( void);
static void delay_15ms( void);
static UWORD ICS2595_Pword( unsigned MHz100);
static void Strobe(void);
static void ICS2595_1bit(char data);
static void InitClk2595( void);
static void Reset_Passthr(UCHAR A);
static WORD IsVRAM(void);
static WORD	DACSubType(void);
static void	ComplexDummyRead(UBYTE *io_base);
static UBYTE	DACComplexRead(UBYTE *io_base);
static void	DACComplexWrite(UBYTE *io_base, UBYTE al);
static WORD CheckDAC(UBYTE *io_base, UBYTE ah, UBYTE bh, UBYTE cl);
static void DAC_read4(void);


ULONG scr_size1;

static WORD	freq_tab[] =
{
	25, 0, 0x10,	28, 0, 0x11,	32, 0, 0x12,	36, 0, 0x13,
	40, 0, 0x04,	45, 0, 0x05,	50, 0, 0x07,	57, 0, 0x01,
	65, 0, 0x0f,	75, 0, 0x0e,	80, 0, 0x0d,
	110,0, 0x0a,	126,0, 0x0a,	135,0, 0x0b
};

static WORD	Min_Freq, Max_Freq, RefFreq = 1432, RefDivider, N_adj;
static WORD	*FreqTable;
static WORD	ClockType;
static WORD	dotClock;
UBYTE	dacType;
static UBYTE hw_flags[4];

/* Bitmaskentabelle zur Zerlegung eines Wertes in einzelne Bitgruppen */
static UBYTE	bitmask[9] = {0x00, 0x01, 0x03, 0x07,
						  0x0f, 0x1f, 0x3f, 0x7f, 0xff};

UCHAR	*io_base;

void	templemon(void) 0x4afc;
UWORD	rolw(UWORD val, UWORD shift) 0xE378;

void outpw(WORD offset, UWORD value)
{
	*(UWORD *)(io_base + offset) = rolw(value, 8);
}

UWORD inpw(WORD offset)
{
	return(rolw(*(UWORD *)(io_base + offset), 8));
}

void outp(WORD offset, UCHAR value)
{
	*(UCHAR *)(io_base + offset) = value;
}

UCHAR inp(WORD offset)
{
	return(*(UCHAR *)(io_base + offset));
}

UBYTE DAC_Type(void)
{
	UBYTE	dacType = inp(DAC_TYPE) & 0x07;
#if !defined(__PANTHER_VRAM) && (defined(__PANTHER) || defined(__MEDUSA) || defined(__HADES))
	if (dacType == 5)
		dacType = 0;
#endif
	return(dacType);
}

RESOLUTION	*VGAOn(RESOLUTION *res)
{
	static WORD	first = 1;
	ULONG	old_val;
	
	if (first)
	{
		WORD	i;
		
#ifdef __PCI
		UCHAR *tmp = NULL;
		LONG	*tmp1 = NULL;
		WORD	*tmp2;
		
		tmp2 = (WORD *)0xA0010002L;
				
		if (*tmp2 == 'XG')
		{		
			tmp = (UCHAR *)0xA0010004L;
			tmp1 = (LONG *)0xA0010010L;
		}

		tmp2 = (WORD *)0xA0020002L;
				
		if (*tmp2 == 'XG')
		{
			tmp = (UCHAR *)0xA0020004L;
			tmp1 = (LONG *)0xA0020010L;
		}

		tmp2 = (WORD *)0xA0040002L;
				
		if (*tmp2 == 'XG')
		{
			tmp = (UCHAR *)0xA0040004L;
			tmp1 = (LONG *)0xA0040010L;
		}

		tmp2 = (WORD *)0xA0080002L;
				
		if (*tmp2 == 'XG')
		{
			tmp = (UCHAR *)0xA0080004L;
			tmp1 = (LONG *)0xA0080010L;
		}
		*tmp = 3;
		*tmp1 = 0L;
#endif

		io_base = IO_BASE;
		InitHardware();							/* Basisinitialisierung */
		FillColors(0);								/* Farbregister lschen */
		first = 0;
	}

	if (res == NULL)								/* Zeiger auf Struktur? */
		res = &standard;							/* nein, Standardauflsung */

	BasicInit(res, io_base);
	InitEngine();

	return(res);									/* fehlerfrei */
}

void InitDataStructure( void)
{
}

void InitM64Reg( void)
{
    static UWORD init_vals[] = {
#if 0
												BUS_CNTL1_16, 0x20F1,
#endif
												BUS_CNTL2_16, 0x890E,
												CRTC_INT_CNTL, 0x0000,
												CRTC_GEN_CNTL, 0x0200,
												GEN_TEST_CNTL, 0x0000,
												(GEN_TEST_CNTL+2), 0x0000,
												MEM_CNTL8, 0x01f0,
												MEM_CNTL8+2, 0,
#if 0
												CRTC_EXT_DISP, 0x0000,
#endif
												SCRATCH_REG0, 0x0000,
												( SCRATCH_REG0 + 2 ), 0x0000,
												SCRATCH_REG1, 0x0000,
												( SCRATCH_REG1 + 2 ), 0x0000,
												DAC_CNTL1, 0x0000,
												CONFIG_CNTL, 0x0011,
												0x0000, 0x0000  };
	int     i;

	outp(BUS_CNTL1_16, 0xf1);
	outp(BUS_CNTL1_16 + 1, 0x20);
	for( i = 0; init_vals[ i] != 0; i += 2)
	{
		outpw(init_vals[i], init_vals[i + 1]);
	}
	outp(CRTC_EXT_DISP, 0x00);
}

void ClrDACCntr(void)
{
	inp(DAC_W_INDEX);
}

UCHAR Passthr_on(void)
{
	UCHAR temp;

	temp = inp( CRTC_EXT_DISP);
	outp( CRTC_EXT_DISP,temp | CRTC_EXT_DISP_EN);

	return( temp);
}

static void Reset_Passthr(UCHAR A)
{
	ClrDACCntr();
	outp( CRTC_EXT_DISP, A);
}

/*********************************************************/
/*							Clock-Chip-Teil							*/
/*********************************************************/
static void Prog_ICS2595( char entry, unsigned MHz100)
{
	UWORD		i, pWord;

	pWord = ICS2595_Pword( MHz100);

	outp( CLOCK_CNTL, 0);
	Strobe();
	outp( CLOCK_CNTL, 1);
	Strobe();

	ICS2595_1bit( 1);
	ICS2595_1bit( 0);                               /* start bits */
	ICS2595_1bit( 0);

	for( i = 0; i < 5; ++i)
	{
		ICS2595_1bit( entry & 1);
		entry >>= 1;
	}

	for( i = 0; i < 8 + 1 + 2 + 2; ++i)
	{
		ICS2595_1bit( pWord & 1);
		pWord >>= 1;
	}
}

static UWORD ICS2595_Pword( unsigned MHz100)
{
	UWORD	divider, ProgramWord;
	ULONG	temp;

	if( MHz100 < Min_Freq) MHz100 = Min_Freq;
	if( MHz100 > Max_Freq) MHz100 = Max_Freq;
	for( divider = 0x03; MHz100 < ( Min_Freq << 3); --divider)
		MHz100 <<= 1;
	if( MHz100 > Max_Freq) MHz100 = Max_Freq;
	temp = ( long unsigned) MHz100;
	temp = ( long unsigned) temp * RefDivider;
	temp += ( RefFreq >> 1);
	temp = temp / RefFreq;
	ProgramWord = ( unsigned )temp;
	if( ProgramWord > N_adj)
		ProgramWord -= N_adj;
	else
		ProgramWord = 0;
	ProgramWord &= 0xff;
	ProgramWord |= ( divider << 9);
	ProgramWord |= 0x1800;                        /*stop bits*/
	return( ProgramWord);
}

void FS2( char data)
{
	outp(CLOCK_CNTL, ( inp(CLOCK_CNTL) & ~0x04 ) | ( data << 2));
}
void FS3( char data)
{
	outp( CLOCK_CNTL, ( inp(CLOCK_CNTL) & ~0x08 ) | ( data << 3));
}
static void Strobe(void)
{
	delay_26us();
	outp( CLOCK_CNTL, inp( CLOCK_CNTL) | CLOCK_STROBE);
}
static void ICS2595_1bit(char data)
{
	FS2( data);
	FS3( 0);
	Strobe();
	FS3( 1);
	Strobe();
}

UCHAR Setup_Prog_Mode( void)
{
	UCHAR	cEntry;
	cEntry = inp( CLOCK_CNTL);
	outp( CLOCK_CNTL, 0);
	return( cEntry);
}

void Restore_Prog_Mode(UCHAR clockCntl)
{
	outp( CLOCK_CNTL, clockCntl | CLOCK_STROBE );
}

static void delay(WORD delayTime)
{
	if (delayTime)
	{
		LONG	endTime;
		delayTime /= 5;
		endTime = *hz200 + delayTime + 2;
		
		while (*hz200 < endTime);
	}
}

static void delay_26us( void)
{
	delay(0);
}

static void delay_1ms( void)
{
	delay(1);
}

static void delay_15ms( void)
{
	delay(15);
}

void ProgramClk(char ClkCntl, UWORD MHz100)
{
	UCHAR	cDisp, cEntry;

	if( ClockType == 1)
	{
		cEntry = Setup_Prog_Mode();
		cDisp = Passthr_on();
		delay_15ms();
		Prog_ICS2595( ClkCntl, MHz100);
		delay_1ms( );
		Reset_Passthr( cDisp);
		Restore_Prog_Mode( cEntry);
	}
	else if (ClockType == 3)
	{
#if 1
#define F(a, b, c) (a<<8)|b|(c<<14)
		static UWORD freq_tab[] =
		{
			14, F(8, 72, 3),	25, F(3, 27, 2),
			27, F(7, 61, 2),	28, F(8, 71, 2),
			30, F(3, 34, 2),	32, F(8, 80, 2),
			34, F(5, 58, 2),	36, F(3, 17, 1),
			38, F(6, 34, 1),	40, F(8, 48, 1),
			44, F(5, 35, 1),
			47, F(6, 44, 1),	48, F(3, 25, 1),
			50, F(3, 27, 1),	52, F(6, 50, 1),
			55, F(7, 61, 1),	57, F(8, 71, 1),
			60, F(3, 34, 1),	63, F(8, 80, 1),
			65, F(8, 83, 1),	68, F(5, 58, 1),
			72, F(3, 17, 0),	75, F(6, 34, 0),
			77, F(6, 35, 0),	80, F(8, 48, 0),
			88, F(5, 35, 0),	90, F(6, 42, 0),
			95, F(3, 25, 0),	100, F(4, 34, 0),
			105, F(7, 58, 0),	108, F(7, 60, 0),
			110, F(7, 61, 0),	118, F(2, 25, 0),
			120, F(3, 34, 0),	131, F(5, 56, 0),
			135, F(5, 38, 0),	999, 0
		};
		int	i;
		
		cEntry = Setup_Prog_Mode();
		cDisp = Passthr_on();
		MHz100 /= 100;
		i = 0;
		while (MHz100 > freq_tab[i])
			i += 2;
		if ((freq_tab[i] - MHz100) >
			 (MHz100 - freq_tab[i-2]))
			i -= 2;

		if (ClkCntl > 0x10)
			ClkCntl = 0x10;
		outp(DAC_CNTL1, 1);
		outp(DAC_W_INDEX, ClkCntl);
		outp(DAC_DATA, (UCHAR)freq_tab[i+1]);
		outp(DAC_DATA, (UCHAR)(freq_tab[i+1]>>8));
		if (ClkCntl < 0x10)
		{
			inp(DAC_MASK);
			inp(DAC_W_INDEX);
			inp(DAC_W_INDEX);
			inp(DAC_W_INDEX);
			inp(DAC_W_INDEX);
			outp(DAC_W_INDEX, 0x80 | ClkCntl);
		}
		outp(DAC_CNTL1, 0);
		Reset_Passthr( cDisp);
		Restore_Prog_Mode( cEntry);
#endif
	}
#if 0
    else  if ( ClockType ==2 ) {    /* if STG1703 DAC */
        cDisp = Passthr_on();
        Prog_STG1703( ClkCntl, MHz100);
        Reset_Passthr( cDisp);
    }
#endif
}

void FixFreq(char ClkCntl)
{
	UCHAR	cDisp, cEntry;
	cEntry = Setup_Prog_Mode();
	cDisp = Passthr_on();

	outp(DAC_CNTL1, 1);

	inp(DAC_MASK);
	inp(DAC_W_INDEX);
	inp(DAC_W_INDEX);
	inp(DAC_W_INDEX);
	inp(DAC_W_INDEX);
	outp(DAC_W_INDEX, ClkCntl);

	outp(DAC_CNTL1, 0);
	Reset_Passthr( cDisp);
	Restore_Prog_Mode( cEntry);
}

static void InitClk2595( void)
{
	static WORD FreqTbl_2595[] = {5035,  5664,  6300,  7200,
											4000,  4490,  4950,  5000,
											0000, 11000, 12600, 13500,
											0000,  8000,  7500,  6500,
											5500,  5500,  5500,  5500
											};
	int     i;
	UCHAR	save, cDisp;

	Min_Freq = 8607 >> 3;
	Max_Freq = 17000;
	RefFreq = 1432;
	RefDivider = 46;
	N_adj = 257;
	FreqTable = FreqTbl_2595;

	save = Setup_Prog_Mode();
	cDisp = Passthr_on();
	delay_15ms();
	for( i = 0; i < 0x10; ++i)
		Prog_ICS2595( i, FreqTable[i]);
	Prog_ICS2595(0x12, FreqTable[0x12]);
	if (dacType == 5)
		ProgramClk(0x13, 5480);
	else
		ProgramClk(0x13, 4700);
	Reset_Passthr( cDisp);
	Restore_Prog_Mode(save);
}

static WORD IsVRAM()
{
	UWORD	ramtype;

	ramtype = ( inp( CONFIG_STAT) >> 3 ) & 0x03;

	if( ( ramtype == 3 ) || ( ramtype ==  1) )
		return(1);
	else
		return(0);
}

WORD TestMemSize(UWORD size, UWORD testlocation)
{
	UWORD	d, data;

/* set memory size */
	outp( MEM_CNTL8, ( inp( MEM_CNTL8) & ~MEM_SIZK_MASK) | ( size & MEM_SIZK_MASK));

/* Write data */
	outpw( TEST_MEM_ADDR, 0x00);
	outpw( TEST_MEM_ADDR + 2, testlocation);

	data = 0x1234;
	outpw( TEST_MEM_DATA, data);

	d = inpw( GEN_TEST_CNTL2);
	d &= ~( GEN_TEST_MODE_MASK | GEN_TEST_MEM_WR_MASK | GEN_TEST_MEM_STROBE);
	d |= ( GEN_TEST_MODE_MEM | GEN_TEST_MEM_WR_WRITE | GEN_TEST_MEM_STROBE);
	outpw( GEN_TEST_CNTL2, d);

/* Read data */
	outpw( TEST_MEM_ADDR, 0x00);
	outpw( TEST_MEM_ADDR + 2, testlocation);

	d = inpw( GEN_TEST_CNTL2);
	d &= ~( GEN_TEST_MODE_MASK | GEN_TEST_MEM_WR_MASK | GEN_TEST_MEM_STROBE);
	d |= ( GEN_TEST_MODE_MEM | GEN_TEST_MEM_WR_READ | GEN_TEST_MEM_STROBE);
	outpw( GEN_TEST_CNTL2, d);

/*compare if they are the same*/
	if( data != inpw( TEST_MEM_DATA))
		return(0);

	return(1);
}

WORD	MemSizeOk(void)
{
	WORD	i;
	UBYTE	*p;
	
	for (i = 0, p = VGA_SCREEN; i < 10; i++)
		*p++ = 0x55;
	for (i = 0, p = VGA_SCREEN; i < 10; i++)
		if (*p++ != 0x55)
			return(0);
	for (i = 0, p = VGA_SCREEN; i < 10; i++)
		*p++ = 0xaa;
	for (i = 0, p = VGA_SCREEN; i < 10; i++)
		if (*p++ != 0xaa)
			return(0);
	
	return(1);
}

WORD	MemSizeOkBank(void)
{
	WORD	i;
	UBYTE	*p;
	
	for (i = 0, p = VGA_SCREEN + 0x200000L; i < 10; i++)
		*p++ = 0x55;
	for (i = 0, p = VGA_SCREEN + 0x200000L; i < 10; i++)
		if (*p++ != 0x55)
			return(0);
	for (i = 0, p = VGA_SCREEN + 0x200000L; i < 10; i++)
		*p++ = 0xaa;
	for (i = 0, p = VGA_SCREEN + 0x200000L; i < 10; i++)
		if (*p++ != 0xaa)
			return(0);
	
	return(1);
}

void MemSizing( void)
{
	UWORD	d, ramtest;

	ramtest = FALSE;

		ramtest = TestMemSize( MEM_SIZE_8M, 8);
		if( !ramtest ) ramtest = TestMemSize( MEM_SIZE_6M, 6);

	if( !ramtest )
		if( !IsVRAM() ) ramtest = TestMemSize( MEM_SIZE_4M, 4);
	if( !ramtest ) ramtest = TestMemSize( MEM_SIZE_2M, 2);
	if( !ramtest ) ramtest = TestMemSize( MEM_SIZE_1M, 1);

	d = inpw( GEN_TEST_CNTL2);
	d &= ~( GEN_TEST_MODE_MASK | GEN_TEST_MEM_WR_MASK | GEN_TEST_MEM_STROBE);

	if (dacType == 5)
	{
											/* +*+*: 4 MByte Speicher */
		outp(MEM_CNTL8, (inp(MEM_CNTL8) & 0xf8) | 3);
		scr_size1 = 0x400000L;
		if (!MemSizeOkBank())
		{
			outp(MEM_CNTL8, (inp(MEM_CNTL8) & 0xf8) | 2);
			scr_size1 = 0x200000L;
		}
	}
	else
	{
		outp(MEM_CNTL8, (inp(MEM_CNTL8) & 0xf8) | 2);
		scr_size1 = 0x200000L;
		if (!MemSizeOk())
		{
			outp(MEM_CNTL8, (inp(MEM_CNTL8) & 0xf8) | 1);
			scr_size1 = 0x100000L;
		}
		hw_flags[2] |= 3;
	}
	outpw( GEN_TEST_CNTL2, d);

	outpw( SCRATCH_REG0, 0x00);
	outpw( SCRATCH_REG0 + 2, 0x00);
	outpw( SCRATCH_REG1, 0x00);
	outpw( SCRATCH_REG1 + 2, 0x00);
}

static void EnableVGA(void)
{
	outp(VGA_CONTROL, 0x16);
	outp(MC_SETUP, 1);
	outp(VGA_CONTROL, 0x0e);
}

static void SetAtiExtended(void)
{
	outp(GRAPH_CONTR, 0x50);
	outp(GRAPH_CONTR + 1, 0xce);
	outp(GRAPH_CONTR, 0x51);
	outp(GRAPH_CONTR + 1, 0x81);
}

static void ScreenOffVGA(void)
{
	outp(SEQU_CONTR, 0x01);
	outp(SEQU_CONTR + 1, 0x20);
}

static void InitVGA(void)
{
	outp(MISC_OUT_W, 0x04);
	inp(INP_STAT_1);
	inp(INP_STAT_1 - 0x20);
	outp(ATTR_CONTR, 0);
	
	ScreenOffVGA();
	
	outp(ATI_EXT, 0xa0);
	outp(ATI_EXT + 1, 0x48);
	
	outp(DAC_MASK_VGA, 0x1d);
	outp(DAC_READ_VGA, 0x00);
	outp(DAC_WRITE_VGA, 0x02);
	
	outp(ATI_EXT, 0xa0);
	outp(ATI_EXT + 1, 0x68);

	outp(DAC_WRITE_VGA, 0x61);
	
	outp(ATI_EXT, 0xa0);
	outp(ATI_EXT + 1, 0x08);
	
	outp(CRTC_EXT_DISP, 0x01);
	outp(DAC_CNTL1, 3);
	outp(DAC_WRITE, 0x61);
	outp(DAC_CNTL1, 2);
	outp(DAC_MASK, 0x1d);
	outp(DAC_READ, 0x80);
	outp(DAC_WRITE, 0x02);
	outp(DAC_CNTL1, 3);
	outp(DAC_WRITE, 0x65);
	outp(DAC_CNTL1, 0);
	
	outp(DAC_CNTL1 + 1, 0x20);
	
	outp(CRTC_EXT_DISP, 0);
	outp(CLOCK_CNTL, 0x4d);
	
	outp(DAC_MASK_VGA, 0xff);
	outp(DAC_WRITE_VGA, 0x00);
	outp(DAC_DATA_VGA, 0x00);
}

static void InitEngine(void)
{
	UBYTE	*reg = (VGA_ADR + 0xbfc00L);

	outpw(BUS_CNTL1_16, inpw(BUS_CNTL1_16) | 0x00a0);
	outp(GEN_TEST_CNTL+1, 0);
	outp(GEN_TEST_CNTL+1, 1);			/* Engine einschalten */
	 
	*(ULONG *)(reg + DP_WRITE_MSK) = 0xffffffffUL;
	*(ULONG *)(reg + CLR_CMP_CNTL) = 0;
}

void InitDAC(void)
{
	dacType = DAC_Type();
	if (dacType == 7)
	{
		UCHAR	cDisp, cEntry;

		cEntry = Setup_Prog_Mode();
		cDisp = Passthr_on();
		inp(DAC_MASK);
		inp(DAC_MASK);
		inp(DAC_MASK);
		if (inp(DAC_MASK) == (UBYTE)192)
			dacType = 0;
		Reset_Passthr( cDisp);
		Restore_Prog_Mode( cEntry);
	}
	if (dacType == 5)
	{
		UCHAR cDisp = Passthr_on();
		SetRS2RS3(DAC_EXT_SEL_RS2);
		outp(DAC_W_INDEX, 0);
		outp(DAC_DATA, 0);
		outp(DAC_DATA, 0);
		outp(DAC_DATA, 0);
		outp(DAC_DATA, 0xff);
		outp(DAC_DATA, 0xff);
		outp(DAC_DATA, 0xff);
		SetRS2RS3(DAC_EXT_SEL_RS3);
		outp(DAC_DATA, 0x80);
		SetRS2RS3(DAC_EXT_SEL_RS3|DAC_EXT_SEL_RS3);
		outp(DAC_W_INDEX, 0xec);
		SetRS2RS3(0);
		Reset_Passthr(cDisp);
	}
}

void InitHardware(void)
{
	EnableVGA();
	
	InitM64Reg();
	InitDAC();
	
	SetAtiExtended();
	
	outp(ATI_EXT, 0xb9);
	outp(ATI_EXT + 1, 0x40);
	outp(SPECIAL_REG, 0);
	EnableVGA();
	InitVGA();
	
	ClockType = 1;
/*	crtc_ext_disp = 3;*/
	
	InitClk2595();

	MemSizing();
	InitDataStructure();
	if (dacType == 7)
	{
		hw_flags[3] = DACSubType();
	}
	else if (dacType == 0)
		ClockType = 3;
/*	PrintData(); */
}

void SetRS2RS3(UCHAR dacRS2RS3)
{
	outp( DAC_CNTL1,
		( inp( DAC_CNTL1) & ~( DAC_EXT_SEL_RS2 | DAC_EXT_SEL_RS3)) | dacRS2RS3);
}

static UCHAR GetMemSize(void)
{
    return( inp( MEM_CNTL8) & 0x07);
}

/******************************************************************************
    A68860_Program
    This routine programs A68860 ramdac.
    INPUT:  colorDepth -
            DispInfo   -
    OUTPUT: None
******************************************************************************/
void A68860_Program(UCHAR colorDepth, UCHAR DispInfo )
{
	typedef struct
	{
		UCHAR	gmode;
		UCHAR	dsetup;
	} A68860_DAC_Table;
	static A68860_DAC_Table	A860_Tab[] =
	{
		{ 0x01, 0x63 },
		{ 0x82, 0x61 },
		{ 0x83, 0x61 },
		{ 0xB0, 0x60 },
		{ 0xB1, 0x60 },
		{ 0xD0, 0x60 },
		{ 0xF3, 0x60 },
		{ 0x80, 0x61 }
	};
	char	d;
	A68860_DAC_Table	*pDacProgTab;
	UCHAR	mask;
	int	ramsize;

	pDacProgTab= &A860_Tab[ colorDepth];
	if( DispInfo & 0x80)  /* if VGA mode */
		pDacProgTab= &A860_Tab[ 7];

	SetRS2RS3( DAC_EXT_SEL_RS2 | DAC_EXT_SEL_RS3);
	outp(DAC_W_INDEX, inp(DAC_W_INDEX) & 0xfd);

	SetRS2RS3( DAC_EXT_SEL_RS3);
	outp( DAC_MASK, 0x1d);

	d = pDacProgTab->gmode;
	if( d & 0x60)  /* if 16bpp or higher */
		if( DispInfo & PITCH_INFO_DAC)
			d |= 0x10;
	outp(DAC_R_INDEX, d);
	outp(DAC_W_INDEX, 0x02);
	SetRS2RS3( DAC_EXT_SEL_RS2 | DAC_EXT_SEL_RS3);
	ramsize = GetMemSize( );
	if( ramsize < MEM_SIZE_1M)
		mask = 4;
	else
		if( ramsize == MEM_SIZE_1M)
			mask = 8;
		else
			mask = 0x0c;
	d = pDacProgTab->dsetup;
	if( DispInfo & PITCH_INFO_DAC)
		d &= 0xfe;

	inp(DAC_W_INDEX);
	outp(DAC_W_INDEX, 0xec);

	SetRS2RS3( 0);
}

static void DAC_read4(void)
{
	inp( DAC_W_INDEX);

	inp( DAC_MASK);
	inp( DAC_MASK);
	inp( DAC_MASK);
	inp( DAC_MASK);
}

static void STG170X_Program(UCHAR colorDepth, UCHAR DispInfo,
									 void *dac_prog_table)
{
	UCHAR	*pDacProgTab;
   WORD	temp;

    pDacProgTab = (UCHAR *) dac_prog_table+colorDepth;

    if(*pDacProgTab==0xff) {
        DAC_read4();
        outp( DAC_MASK, 0x19);
        return;
    }
									/* bei 1280x1024 */
    if(DispInfo & PITCH_INFO_MUX2)
        pDacProgTab=(UCHAR *) dac_prog_table+STG170X_MUX8;

    DAC_read4();
									/* bei 1280x1024 */
									/* bei 640x480x32 */
									/* bei 800x600x24 */
    if(DispInfo & PITCH_INFO_DAC)
        outp(DAC_MASK,0x1a);
    else
        outp(DAC_MASK,0x18);

    DAC_read4();

    inp( DAC_MASK);
    outp( DAC_MASK, 3);
    outp( DAC_MASK, 0);

    outp( DAC_MASK, *pDacProgTab);

    outp( DAC_MASK, *pDacProgTab);

    if( *pDacProgTab == 5) {
        outp( DAC_MASK, 2);
        return;
    }

    if (*pDacProgTab==9)
    {
		temp=1;
		if (dotClock >= 3200)
		{
			temp++;
			if (dotClock >= 6400)
				temp++;
		}
		outp(DAC_MASK, temp);
	}
}

static void ATT21C498_Program(UCHAR colorDepth, UCHAR DispInfo,
									 void *dac_prog_table)
{
    UCHAR *pDacProgTab, dacMask;

    pDacProgTab = (UCHAR *) dac_prog_table+colorDepth;

    if(DispInfo & PITCH_INFO_MUX2)
        pDacProgTab=(UCHAR *) dac_prog_table+7;

    DAC_read4();

    dacMask = *pDacProgTab;
    if(DispInfo & PITCH_INFO_DAC)
        dacMask |= 2;
    outp( DAC_MASK, dacMask);

}

static void STG1702_Program(UCHAR colorDepth, UCHAR DispInfo)
{
	static  UCHAR STG1702_Tab[] = {0xff, 0, 0, 2, 3, 9, 0, 5};
	static  UCHAR ATT21C498_Tab[]={0x0c, 4, 4, 0x16, 0x36, 0xe6, 0xe6, 0x24};
	static  UCHAR CH8398_Tab[]   ={0x0c, 4, 4, 0x16, 0x36, 0xb6, 0xb6, 0x24};

	if (colorDepth == 6)
		colorDepth = 5;
	if (dacType != 7)
		ATT21C498_Program(colorDepth, DispInfo, CH8398_Tab);
	else if (hw_flags[3] == 2)
		ATT21C498_Program(colorDepth, DispInfo, ATT21C498_Tab);
	else
	{
		STG170X_Program(colorDepth, DispInfo, STG1702_Tab);
	}
}

void DAC_Programming(UCHAR colorDepth, UCHAR DispInfo)
{
	UCHAR	temp;
	UCHAR	crtcGenCntl;

	if( DispInfo & 1)
	{
		crtcGenCntl = inp(CRTC_GEN_CNTL) & ~CRTC_PIX_BY_2_EN;
		outp( CRTC_GEN_CNTL, crtcGenCntl);
	outp( CRTC_EXT_DISP, CRTC_EXT_DISP_EN | CRTC_EXT_EN);
	}
	else
		outp( CRTC_EXT_DISP, 0);

	temp = Passthr_on();

	if (dacType == 5)
		A68860_Program(colorDepth, DispInfo);
	else
		STG1702_Program(colorDepth, DispInfo);
	Reset_Passthr( temp);

	if( DispInfo & 1)
	{
		crtcGenCntl = inp(CRTC_GEN_CNTL) & ~CRTC_PIX_BY_2_EN & ~CRTC_BLANK;
									/* bei 1280x1024 */
		if ((DispInfo & PITCH_INFO_MUX2) && 1)
			crtcGenCntl |= CRTC_PIX_BY_2_EN;
		outp( CRTC_GEN_CNTL, crtcGenCntl);
	}
}

static VOID	CalcResValues(RESOLUTION *r, ATI_RES *res)
{
	WORD	help;
	WORD	h_sync_end, v_sync_end;
	
	res->freq = ConvFreqInt(r->freq);	/* Index der Frequenz bestimmen */
	res->flags = 0;
	if (!(r->r_3c2 & 0x40))					/* hor. Polaritt aktiv? */
		res->flags |= HOR_POLARITY;		/* ja */
	if (!(r->r_3c2 & 0x80))					/* ver. Polaritt aktiv? */
		res->flags |= VER_POLARITY;		/* ja */
	if (r->extended[2] & 0x80)				/* Interlaced? */
		res->flags |= INTERLACED;			/* ja */
	if (r->r_3d4[9] & 0x80)					/* Doublescan? */
		res->flags |= DOUBLESCAN;			/* ja */

	res->mode = r->mode;						/* Auflsungsmodus bernehmen */
	res->planes = r->planes;				/* Anzahl Planes bernehmen */

	res->h_total = ReadBit(r->r_3d4[0x00], 8, 0, 0) +
					   (((UWORD)r->low_res & 4)<<6) + 5;
	if (r->flags & HTOTAL_9)
		res->h_total += 512;
	res->h_disp	= ReadBit(r->r_3d4[0x01], 8, 0, 0) +
						(((UWORD)r->low_res & 8)<<5) + 1;
	if (r->flags & HDEND_9)
		res->h_disp += 512;
	res->h_sync_strt = ReadBit(r->r_3d4[0x04], 8, 0, 0) +
						(((UWORD)r->low_res & 32)<<3);
	if (r->flags & HSSTART_9)
		res->h_sync_strt += 512;
	if (res->h_sync_strt < res->h_disp / 2)
		res->h_sync_strt += 256;
	h_sync_end = ReadBit(r->r_3d4[0x05], 5, 0, 0);
	res->h_sync_wid = (h_sync_end - (res->h_sync_strt & 0x3f));
	if (res->h_sync_wid < 0)
		res->h_sync_wid += 0x40;

	res->v_total= (ReadBit(r->r_3d4[0x06], 8, 0, 0) |
					  ReadBit(r->r_3d4[0x07], 1, 0, 8) |
					  ReadBit(r->r_3d4[0x07], 1, 5, 9) |
					  ReadBit(r->extended[2], 1, 1, 10)) + 2;
	res->v_disp	= (ReadBit(r->r_3d4[0x12], 8, 0, 0) |
					  ReadBit(r->r_3d4[0x07], 1, 1, 8) |
					  ReadBit(r->r_3d4[0x07], 1, 6, 9) |
					  ReadBit(r->extended[2], 1, 2, 10)) + 1;
	res->v_sync_strt=ReadBit(r->r_3d4[0x10], 8, 0, 0) |
					  ReadBit(r->r_3d4[0x07], 1, 2, 8) |
					  ReadBit(r->r_3d4[0x07], 1, 7, 9) |
					  ReadBit(r->extended[2], 1, 3, 10);
	v_sync_end	= ReadBit(r->r_3d4[0x11], 4, 0, 0);
	
	if (r->planes <= 4)
	{
		r->bypl = (r->max_x + 1)>>3;
		r->r_3d4[0x13] = r->bypl>>1;
	}
	else
	{
		help = r->planes / 8;				/* Teiler fr hor. Werte */
		if (r->planes < 32)
		{

			res->h_total /= help;
			res->h_disp /= help;
			res->h_sync_strt /= help;
			res->h_sync_wid /= help;
		}
		if (r->planes <= 24)
		{
			res->pitch = (UWORD)(((r->max_x + 128)>>3) & 0xfff0);
			r->bypl = res->pitch * r->planes;
			if ((LONG)r->bypl * (LONG)(r->max_y + 1) > scr_size1)
			{
				r->bypl = (r->max_x + 1) * help;
				res->pitch = (UWORD)((r->max_x + 1)>>3);
			}
		}
		else
		{
			r->bypl = (r->max_x + 1) * help;
			res->pitch = (UWORD)((r->max_x + 1)>>3);
		}
	}
	if (res->flags & DOUBLESCAN)
	{
		res->v_total >>= 1;
		res->v_disp >>= 1;
		res->v_sync_strt >>= 1;
		v_sync_end >>= 1;
	}
	help = res->v_sync_strt & 0x0f;		/* relevanter Teil */
	v_sync_end -= help;						/* Breite des Sync-Signals */
	if (v_sync_end < 0)
		v_sync_end += 0x10;
	res->v_sync_wid = v_sync_end;

	res->h_total--;
	res->h_disp--;
	res->h_sync_strt--;
	res->v_total--;
	res->v_disp--;
	res->v_sync_strt--;

	if (res->h_sync_wid > 0x1f)
		res->h_sync_wid = 0x1f;
}


static void	FillColors(char value)
{
	int	i;
	
	*(io_base + DAC_WRITE) = 0;				/* bei Eintrag 0 beginnen */
	
	for (i = 0; i < 768; i++)					/* alle Eintrge beschreiben */
	{
		*(io_base + DAC_DATA) = value;
	}
}

static UWORD FIFO_depth(void)
{
	UWORD	temp, depth, size;
	static WORD div[ 8] = { 6, 6, 12, 12, 12, 12, 12, 12};
	/* JY 04/07/94 added */
	static UCHAR fifoVals[][6]={{0x02, 0x0c, 0x02, 0x06, 0x02, 0x0e}, /* 8bpp */
										 {0x02, 0x0e, 0x02, 0x0e, 0x02, 0x0e}, /* 16bpp */
										 {0x02, 0x0e, 0x02, 0x0e, 0x02, 0x0e}, /* 24bpp */
										 {0x02, 0x0e, 0x02, 0x0e, 0x02, 0x0e}, /* 24bpp */
										 {0x02, 0x0e, 0x02, 0x0e, 0x02, 0x0e}};/* 32bpp */

	temp = dotClock / 100;
	depth = inp(CRTC_PIX_WIDTH) & 7;
	switch (depth)
	{
		case PIX_WIDTH555:
		case PIX_WIDTH565:
			temp <<= 1;
			break;
		case PIX_WIDTH24:
			temp = (temp << 1) + temp;
			break;
		case PIX_WIDTH32:
			temp <<= 2;
			break;
		default:
			break;
	}
	size = GetMemSize();
	temp /= div[size];
	size <<= 1;
	depth -= 2;
	/* JY 04/11/94  added */
	if (temp < fifoVals[depth][size])
		temp = fifoVals[depth][size];
    if (temp > fifoVals[depth][size + 1])
		temp = fifoVals[depth][size + 1];
	return(temp);
}

void	BasicInit(RESOLUTION *r, UCHAR *io_base)
{
	ATI_RES	res;
	UBYTE		i;

	CalcResValues(r, &res);					/* Auflsungswerte berechnen */

#ifdef __PCI
		if (dacType == 5)
			ProgramClk(0x13, 5400);
		else
			ProgramClk(0x13, 4700);
#else
		ProgramClk(0x13, 4000);
#endif
		outpw(CONFIG_CNTL, 0x0011);
		DAC_Programming(2, 1);

		dotClock = r->freq * 100;
		ProgramClk( CX_CLK, dotClock);
		res.freq = CX_CLK;
		
		outp(CRTC_EXT_DISP, 2);
													/* VGA-Passthrough ein */
		EnableVGA();
		*(io_base + VIDEO_ENABLE) = 1;	/* VGA On */
		
		SetAtiExtended();
		
		r->r_3d4[0x0d] = 0;
		r->r_3d4[0x0c] = 0;
		
		if (r->planes == 1)					/* nur eine Plane? */
		{
			sequ_values[1] = 0x01;			/* Zugriff nur auf eine Plane */
			attr_values[18] = 1;				/* nur eine Plane fr Farbbildung */
		}
		else										/* ansonsten alle Planes benutzen */
		{
			sequ_values[1] = 0x0F;			/* Zugriff nur auf eine Plane */
			attr_values[18] = 0x0F;			/* nur eine Plane fr Farbbildung */
		}
	
		sequ_values[3] &= 0xF7;				/* Chain-4 ausschalten */
		graph_values[5] &= 0xBF;			/* 256-Farbmodus */
	
		sequ_values[0] &= 0xF7;
		if (r->low_res & 1)					/* Low Res? */
			sequ_values[0] |= 8;
		
		attr_values[0x16] &= 0x4F;			/* High-Resolution ausschalten */
		sequ_values[5] &= 0xBE;				/* MCLK/2 und /4 ausschalten */

		attr_values[0x11] = 255;			/* Rahmen auf 255 */
	
		SeqReset(io_base);					/* Sequenzer resetten */
													/* Sequenzer initialisieren */
		SetValues(io_base, SEQU_CONTR, sequ_values, 1, 4);
	
		*(io_base + MISC_OUT_W) = r->r_3c2;	/* Misc. Output setzen */
		
													/* CRT-Controller intialisieren */
		*(io_base + CRT_CONTR) = 0x11;
		*(io_base + CRT_CONTR+1) &= 0x7F;
		
		r->r_3d4[7] |= 0x10;
		SetValues(io_base, CRT_CONTR, r->r_3d4, 0, 0x19);
		*(io_base + ATI_EXT) = 0x86;		/* verschiedenes */
		*(io_base + ATI_EXT + 1) = 0x7a;
		*(io_base + ATI_EXT) = 0xa3;		/* Display Buffer Start Bit 17 */
		*(io_base + ATI_EXT + 1) = 0;
		*(io_base + ATI_EXT) = 0xad;		/* Display Buffer Start Bit 18/19 */
		*(io_base + ATI_EXT + 1) = 0;
		*(io_base + ATI_EXT) = 0xae;		/* verschiedenes */
		*(io_base + ATI_EXT + 1) = 0x00;
		*(io_base + ATI_EXT) = 0xb0;		/* Display Buffer Start Bit 16 */
		*(io_base + ATI_EXT + 1) = 8;
		for (i = 0xb1; i <= 0xb5; i++)
		{
			*(io_base + ATI_EXT) = i;		/* verschiedenes */
			*(io_base + ATI_EXT + 1) = 0;
		}
		*(io_base + ATI_EXT) = 0xb6;		/* verschiedenes */
		*(io_base + ATI_EXT + 1) = 0x01;
		*(io_base + ATI_EXT) = 0xb8;		/* verschiedenes */
		*(io_base + ATI_EXT + 1) = 0;
		*(io_base + ATI_EXT) = 0xbd;		/* verschiedenes */
		*(io_base + ATI_EXT + 1) = 4;
		*(io_base + ATI_EXT) = 0xbe;		/* verschiedenes */
		if (res.flags & INTERLACED)
			*(io_base + ATI_EXT + 1) = 0x0a;
		else
			*(io_base + ATI_EXT + 1) = 8;
		*(io_base + ATI_EXT) = 0xbf;		/* verschiedenes */
		*(io_base + ATI_EXT + 1) = 1;
													/* Graphik-Controller initialisieren */
		SetValues(io_base, GRAPH_CONTR, graph_values, 0, 9);
		SetAttribute(io_base);				/* Attribut-Controller initialisieren */

		*(io_base + DAC_MASK) = 0xFF;
													/* Frequenz setzen */
		SetFreq(res.freq, r->extended[3] & 0x80);

	if (dacType == 0)
		FixFreq(CX_CLK);
}

/******************************************************************/
/* Name: ReadBit																	*/
/* Funktion: kopiert Bits aus einem Byte in ein Zielwort				*/
/* Parameter:																		*/
/*		int value: Quellwort														*/
/*		int count: Anzahl zu kopierender Bits								*/
/*		char sbit: Startbit in Quelle											*/
/*		char dbit: Startbit in Ziel											*/
/* Funktionswert:																	*/
/*		int: Zielwort																*/
/******************************************************************/
static int	ReadBit(UINT value, int count, char sbit, char dbit)
{
	if (sbit <= dbit)								/* nach links shiften */
		value <<= (dbit - sbit);
	else
		value >>= (sbit - dbit);				/* nach rechts shiften */
	
	value &= bitmask[count]<<dbit;			/* relevantes Bit isolieren */
	
	return(value);
}


static void	SetAttribute(UBYTE *io_base)
{
	char	i;
	UBYTE	*data = attr_values;					/* Zeiger auf Attributwerte */
	
	i = *(io_base + INP_STAT_1);				/* Dummy-Read */
	for (i = 0; i <= 0x14; i++)				/* alle Register setzen */
	{
		*(io_base + ATTR_CONTR) = i;			/* Index schreiben */
		*(io_base + ATTR_CONTR) = *data++;	/* Daten schreiben */
	}
	*(io_base + ATTR_CONTR) = 0x20;			/* DAC freigeben */
}


/******************************************************************/
/* Name: WriteBit																	*/
/* Funktion: kopiert Bits aus einem Wort in ein Zielbyte				*/
/* Parameter:																		*/
/*		int value: Quellwort														*/
/*		char old_value: alter Wert des Zielbytes							*/
/*		int count: Anzahl zu kopierender Bits								*/
/*		char sbit: Startbit in Quelle											*/
/*		char dbit: Startbit in Ziel											*/
/* Funktionswert:																	*/
/*		char: modifiziertes Byte												*/
/******************************************************************/
static char	WriteBit(UINT value, UCHAR old_value,
							int count, char sbit, char dbit)
{
	if (sbit <= dbit)								/* nach rechts shiften */
		value >>= (dbit - sbit);
	else
		value <<= (sbit - dbit);				/* nach links shiften */
	
	old_value = (value & (bitmask[count]<<sbit)) |
					(old_value & ~(bitmask[count]<<sbit));
	
	return(old_value);
}

void SetReg16(UBYTE *adr, UWORD value)
{
	*(UWORD *)adr = (value<<8) + (value>>8);
}

static UWORD GetReg16(UBYTE *adr)
{
	return(((UWORD)adr[1]<<8) + (UWORD)adr[0]);
}

static void	SetFreq(WORD freq, WORD flag)
{
	freq = 3*freq + 2;
	
	*(io_base + SEQU_CONTR) = 0;		/* Sequenzer Reset */
	*(io_base + SEQU_CONTR + 1) = 1;
	
	*(io_base + ATI_EXT) = 0xb9;		/* Frequenz-Bit 2 */
	*(io_base + ATI_EXT + 1) = 0x40 | ((freq_tab[freq] & 4)>>1);
	*(io_base + ATI_EXT) = 0xb8;		/* Frequenzteiler */
	*(io_base + ATI_EXT + 1) = (freq_tab[freq] & 0x10)<<2;
	*(io_base + ATI_EXT) = 0xbe;		/* Frequenz-Bit 3 */
	*(io_base + ATI_EXT + 1) = ((freq_tab[freq] & 8)<<1) |
										(UBYTE)((flag != 0)*2);
	*(io_base + MISC_OUT_W) = (*(io_base + MISC_OUT_R) & 0xf3) |
										((freq_tab[freq] & 3)<<2);

	*(io_base + SEQU_CONTR) = 0;		/* Sequenzer Reset aus */
	*(io_base + SEQU_CONTR + 1) = 3;
}

static void DummyRead3C8(UBYTE *io_base)
{
	char	j;
	
	j = *(io_base + DAC_WRITE);
}

static UBYTE DummyRead3C6(UBYTE *io_base)
{
	UBYTE	i, j;
	int	k;
	
	for (i = 0; i < 4; i++)
	{
		for (k = 0; k < 1000; k++);
		j = *(io_base + DAC_MASK);
	}
	return(j);
}

static WORD	DACSubType(void)
{
	WORD	dac_type = 0;
	UBYTE	al;
	UBYTE	bl, bh;
	UBYTE cl, ch;
	UBYTE	*io_base = IO_BASE;

	cl = DACComplexRead(io_base);
	bh = *(io_base + DAC_MASK);
	
	ch = ~cl;
	*(io_base + DAC_MASK) = ch;
	
	ComplexDummyRead(io_base);
	al = *(io_base + DAC_MASK);
	bl = 0;
	if (al == ch)
		DummyRead3C8(io_base);
	else
	{
		DACComplexWrite(io_base, 0xe0);
		ComplexDummyRead(io_base);
		al = *(io_base + DAC_MASK) & 0xe0;
		bl = 1;
		if (al == 0xe0)
		{
			DACComplexWrite(io_base, 0x60);
			ComplexDummyRead(io_base);
			al = *(io_base + DAC_MASK) & 0xe0;
			bl = 3;
			DACComplexWrite(io_base, cl);
			if (al)
			{
				ComplexDummyRead(io_base);
				al = *(io_base + DAC_MASK);
				al = *(io_base + DAC_MASK);
				if (al == ch)
				{
					DACComplexWrite(io_base, 0xff);
					ComplexDummyRead(io_base);
					al = *(io_base + DAC_MASK);
					bl = 4;
					if (al & 0xf0)
						bl = 3;
					DACComplexWrite(io_base, al);
				}
				else
				{
					al = DummyRead3C6(io_base);
					bl = 6;
					if (al != ch)
					{
						DummyRead3C8(io_base);
						*(io_base + DAC_MASK) = 0xff;
						al = DummyRead3C6(io_base);
						bl = 7;
						if (al != 0x8e)
						{
							bl = 8;
							if ((al != 0x82) && (al != 0x44))
							{
								al = (UBYTE)CheckDAC(io_base,
															0xef, bh, cl);
								bl = 5;
								if (!al)
								{
									al = (UBYTE)CheckDAC(io_base,
																0xfb, bh, cl);
									bl = 9;
									if (!al)
										bl = 2;
									DACComplexWrite(io_base, cl);
								}
								else
									DACComplexWrite(io_base, cl);
							}
							else
								DACComplexWrite(io_base, cl);
						}
						else
							DACComplexWrite(io_base, cl);
					}
					else
						DACComplexWrite(io_base, cl);
				}
			}
		}
		else
			DACComplexWrite(io_base, cl);
	}
	*(io_base + DAC_MASK) = al;
	dac_type = bl;

	*(io_base + DAC_MASK) = 0xff;
	return(dac_type);
}

static void	ComplexDummyRead(UBYTE *io_base)
{
	DummyRead3C8(io_base);
	DummyRead3C6(io_base);
}

static UBYTE	DACComplexRead(UBYTE *io_base)
{
	UBYTE	al;
	
	ComplexDummyRead(io_base);
	
	al = *(io_base + DAC_MASK);
	
	DummyRead3C8(io_base);
	return(al);
}

static void	DACComplexWrite(UBYTE *io_base, UBYTE al)
{
	ComplexDummyRead(io_base);
	
	*(io_base + DAC_MASK) = al;
	
	DummyRead3C8(io_base);
}

static WORD CheckDAC(UBYTE *io_base, UBYTE ah, UBYTE bh, UBYTE cl)
{
	UBYTE	al;
	
	DummyRead3C8(io_base);
	
	*(io_base + DAC_MASK) = ah & bh;
	
	ComplexDummyRead(io_base);
	
	*(io_base + DAC_MASK) = ~ah | cl;
	
	al = *(io_base + DAC_MASK);
	
	*(io_base + DAC_MASK) = ah & al;
	
	return((!ah & al) != 0);
}
