#ifndef __XCB
#define __XCB

#include	<portab.h>

/******************************************************************/
/* Es folge eine Beschreibung der XCB-Struktur. Diese Struktur		*/
/* wird vom EMULATOR.PRG angelegt, wenn dieses eine NOVA-Karte		*/
/* im Rechner findet.															*/
/* Zu diesem Zweck legt EMULATOR.PRG einen Cookie mit der Kennung	*/
/* 'NOVA' an. Der Wert dieses Cookies zeigt auf die XCB-Struktur.	*/
/* Alle Werte dieser Struktur sind, wenn nicht anders angegeben,	*/
/* nur zum Lesen gedacht.														*/
/******************************************************************/


typedef struct
{
	LONG	version;							/* Versionsnummer */
												/* diese Versionsnummer ist */
												/* unabhngig von der VDI-Version */
												/* folgende vier Werte sollten */
												/* nur Menu-Programme wie Menu, Xmenu */
												/* verwenden: */
	UBYTE	resolution;						/* Auflsungsindex */
	UBYTE	blnk_time;						/* Bildschirmdunkelzeit */
	UBYTE	ms_speed;						/* Mausgeschwindigkeit */
	BYTE	old_res;							/* ursprngliche Auflsung */
												/* Sonderbedeutung fr Bit 7: */
												/* 1: schneller 8-Bit-Text */

												/* Auflsungsumschaltung */
												/* fll_ofst sollt 0UL sein */
	VOID	(*p_chres)(RESOLUTION *res, ULONG fll_ofst);
	WORD	mode;								/* Auflsungsmodus, momentan sind */
												/* folgende Werte definiert: */
												/* 0: 16 Farben */
												/* 1: 2 Farben */
												/* 2: 256 Farben */
												/* 3: 32768 Farben (Intelformat) */
												/* 4: 65536 Farben (Intelformat) */
												/*	5: 16.7 Mio. (24 Bit BGR) */
												/* 6: 16.7 Mio. (32 Bit RGBx) */
	WORD	bypl;								/* Bytes pro Bildschirmzeile, */
												/* dieser Wert mu nicht gleich */
												/* Anzahl x-Pixel * Bytes pro Pixel sein! */
	WORD	planes;							/* Anzahl Bildschirmplanes: */
												/* 1, 4, 8, 16, 24 oder 32 */
	WORD	colors;							/* Anzahl Farben, nicht benutzen */
	WORD	hc;								/* Hardcopymodus: */
												/* 0: 1 Bildpxl. = 1x1 Druckpxl. */
												/* 1: 1 Bildpxl. = 2x2 Druckpxl. */
												/* 2: 1 Bildpxl. = 4x4 Druckpxl. */
	WORD	max_x, max_y;					/* Bildschirmauflsung */
												/* fr virtuelle Bildschirm- */
												/* verwaltung: */
	WORD	rmn_x, rmx_x;					/* physikalisch auf dem Monitor */
	WORD	rmn_y, rmx_y;					/* dargestellter Bereich */
												/* folgende 4 Werte drfen verndert */
												/* werden. Dabei ist aber zu */
												/* bercksichtigen, da die Werte */
												/* sinnvoll bleiben, d.h.: */
												/* v_top + v_bottom < rmx_y - rmn_y */
												/* v_left + v_right < rmx_x - rmn_x */
	WORD	v_top, v_bottom,				/* Rnder fr virt. Speicherverwaltung */
			v_left, v_right;
												/* Zeiger auf Routine zum Farben */
												/* setzen. Index ist der Farb- */
												/* index, colors der Zeiger auf */
												/* das 3-Byte-Array mit den */
												/* Farbwerten */
	VOID	(*p_setcol)(WORD index, UBYTE *colors);
												/* virtuelle Bildschirmverwaltung: */
												/* es wird getestet, ob der Punkt */
												/* (x;y) sich im dargestellten */
												/* Bildabschnitt befindet */
												/* falls nicht, wird der darge- */
												/* stellte Abschnitt so verschoben, */
												/* da der Punkt gerade sichtbar */
												/* wird. Hierbei werden v_top, ... */
												/* bercksichtigt. */
	VOID	(*chng_vrt)(WORD x, WORD y);
												/* XBIOS-Routinen fr Graphikkarte */
												/* installieren/abschalten */
												/* on != 0: installieren */
												/* inst_xbios ist nur fr Menu-Prog. */
												/* wie MENU.PRG und XMENU.PRG */
												/* gedacht. */
	VOID	(*inst_xbios)(WORD on);
												/* Bild ein-/ausschalten */
	VOID	(*pic_on)(WORD on);			/* 0: Bild aus-, 1: Bild einschalten */
												/* Bildlage verndern */
												/* res: Auflsungsstruktur */
												/* direction = 0: hor. Lage ndern */
												/* direction != 0: ver. Lage ndern */
												/* offset: Anzahl Einheiten, um die */
												/*        verschoben werden soll */
												/*        bei links oder oben positiv */
	VOID	(*chng_pos)(RESOLUTION *res, WORD direction, WORD offset);
	VOID	(*p_setscr)(VOID *adr);		/* physikalische Bildschirmadresse */
												/* umsetzen. Die neue Bildschirm- */
												/* adresse mu im Speicher der */
												/* Grafikkarte liegen! */
	VOID	*base;							/* Adresse von Bildschirmseite 0 */
	VOID	*scr_base;						/* Adresse des Bildschirmspeichers */
	UWORD	scrn_cnt;						/* Anzahl mglicher phys. Bildschirm- */
												/* seiten */
	LONG	scrn_sze;						/* Gre eines Bildschirms in Bytes */
	UBYTE	*reg_base;						/* Zeiger auf I/O-Adressen, nie benutzen */
	VOID	(*p_vsync)(VOID);				/* wartet auf vsync */

	BYTE	name[36];						/* Name der aktuellen Auflsung */
												/* folgende Variablen sind erst */
												/* ab Version 1.01 definiert: */
	ULONG	mem_size;						/* Gre des Bildschirmspeichers in Byte */
												/* ab Version 1.20: */
	BYTE	hw_flags[8];					/* 4 kartenabhngige Werte */
												/* Byte 0: Hardwareuntersttzung fr 24/32-Bit */
												/* Byte 1: 24-Bit-Aufbau ist RGB */
												/* Byte 2: Bit 0 != 0: kein Hw-Cursor bei 24-Bit */
												/*         Bit 1 != 0: kein Hw-Cursor bei 32-Bit */
												/* Byte 3: DAC-Subtype */
	WORD	cpu;								/* CPU-Type: 0/10/20/30/40 */
												/* Setzen der Gamma-Kurve */
												/* ist set_gamma == NULL, wird */
												/* dieses Feature nicht untersttzt */
	void	(*set_gamma)(UBYTE *gamma);
	WORD	gamma_able;						/* Flag, ob Gammakorrektur mglich */
	WORD	gamma_rgb[3];
} XCB;

#endif